from pynput import keyboard


supported_control_keys = {
                keyboard.Key.f1: '__F1__',
                keyboard.Key.f2: '__F2__',
                keyboard.Key.f3: '__F3__',
                keyboard.Key.f4: '__F4__',
                keyboard.Key.f5: '__F5__',
                keyboard.Key.f6: '__F6__',
                keyboard.Key.f7: '__F7__',
                keyboard.Key.f8: '__F8__',
                keyboard.Key.f9: '__F9__',
                keyboard.Key.f10: '__F10__',
                keyboard.Key.ctrl_l: '__CTRL_L__',
                keyboard.Key.ctrl_r: '__CTRL_R__',
                keyboard.Key.enter: '__ENTER__',
                keyboard.Key.backspace: '__BACKSPACE__',
                keyboard.Key.esc: '__ESC__',
                keyboard.Key.delete: '__DELETE__',
                keyboard.Key.home: '__HOME__',
                keyboard.Key.page_up: '__PAGE_UP__',
                keyboard.Key.end: '__END__',
                keyboard.Key.page_down: '__PAGE_DOWN__',
                keyboard.Key.insert: '__INSERT__',
                keyboard.Key.up: '__UP__',
                keyboard.Key.down: '__DOWN__',
                keyboard.Key.left: '__LEFT__',
                keyboard.Key.right: '__RIGHT__',
                keyboard.Key.tab: '__TAB__'
}